import type { Equatable } from "@siteimprove/alfa-equatable";
import type { Flags } from "@siteimprove/alfa-flags";
import type { Hash, Hashable } from "@siteimprove/alfa-hash";
import type { Predicate } from "@siteimprove/alfa-predicate";
import { Option } from "@siteimprove/alfa-option";
import { Refinement } from "@siteimprove/alfa-refinement";
import { Sequence } from "@siteimprove/alfa-sequence";
import * as json from "@siteimprove/alfa-json";
/**
 * Model for n-ary trees with some traversal flags.
 *
 * In order to have a parent pointers, nodes are allowed to attach themselves
 * to a parent node. To prevent mutation of an existing tree, the child is then
 * "frozen".
 *
 * Since it is not possible to add children after node creation, and it is not
 * possible to re-attach an already attached node, this means that the trees are
 * effectively downward frozen. In turn, this allows a bunch of optimisation
 * since any traversal function that does not look upward can be cached.
 *
 * The full tree (all nodes) must accept the same set of traversal flags, but
 * the node type is not constrained.
 *
 * @public
 */
export declare abstract class Node<F extends Flags.allFlags, T extends string = string> implements Iterable<Node<F>>, Equatable, Hashable, json.Serializable<Node.JSON<T>> {
    protected readonly _children: Array<Node<F>>;
    protected _parent: Option<Node<F>>;
    protected readonly _type: T;
    private readonly _externalId;
    private readonly _extraData;
    private readonly _serializationId;
    /**
     * Whether the node is frozen.
     *
     * @remarks
     * As nodes are initialized without a parent and possibly attached to a parent
     * after construction, this makes hierarchies of nodes mutable. That is, a
     * node without a parent node may be assigned one by being passed as a child
     * to a parent node. When this happens, a node becomes frozen. Nodes can also
     * become frozen before being assigned a parent by using the `Node#freeze()`
     * method.
     */
    protected _frozen: boolean;
    protected constructor(children: Array<Node<F>>, type: T, externalId?: string, serializationId?: string, extraData?: any);
    get type(): T;
    get externalId(): string | undefined;
    get extraData(): any;
    get serializationId(): string;
    get frozen(): boolean;
    /**
     * Freeze the node. This prevents further expansion of the node hierarchy,
     * meaning that the node can no longer be passed as a child to a parent node.
     */
    freeze(): this;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-parent}
     */
    parent(options?: Flags<F>): Option<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-parent}
     */
    isParentOf(node: Node<F>, options?: Flags<F>): boolean;
    private _lastKnownRoot;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-root}
     */
    root(options?: Flags<F>): Node<F>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-root}
     */
    isRootOf(node: Node<F>, options?: Flags<F>): boolean;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-child}
     */
    children(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-child}
     */
    isChildOf(node: Node<F>, options?: Flags<F>): boolean;
    private _descendants;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-descendant}
     */
    descendants(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-descendant}
     */
    isDescendantOf(node: Node<F>, options?: Flags<F>): boolean;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-inclusive-descendant}
     */
    inclusiveDescendants(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-inclusive-descendant}
     */
    isInclusiveDescendantsOf(node: Node<F>, options?: Flags<F>): boolean;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-ancestor}
     */
    ancestors(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-ancestor}
     */
    isAncestorOf(node: Node<F>, options?: Flags<F>): boolean;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-inclusive-ancestor}
     */
    inclusiveAncestors(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-inclusive-ancestor}
     */
    isInclusiveAncestorOf(node: Node<F>, options?: Flags<F>): boolean;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-sibling}
     */
    siblings(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-sibling}
     */
    isSiblingOf(node: Node<F>, options?: Flags<F>): boolean;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-inclusive-sibling}
     */
    inclusiveSiblings(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-inclusive-sibling}
     */
    isInclusiveSiblingOf(node: Node<F>, options?: Flags<F>): boolean;
    private _preceding;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-preceding}
     */
    preceding(options?: Flags<F>): Sequence<Node<F>>;
    private _following;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-following}
     */
    following(options?: Flags<F>): Sequence<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-first-child}
     */
    first(options?: Flags<F>): Option<Node<F>>;
    private _last;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-last-child}
     */
    last(options?: Flags<F>): Option<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-previous-sibling}
     */
    previous(options?: Flags<F>): Option<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-next-sibling}
     */
    next(options?: Flags<F>): Option<Node<F>>;
    /**
     * {@link https://dom.spec.whatwg.org/#concept-tree-index}
     */
    index(options?: Flags<F>): number;
    /**
     * {@link https://dom.spec.whatwg.org/#dom-element-closest}
     */
    closest<T extends Node<F>>(refinement: Refinement<Node<F>, T>, options?: Flags<F>): Option<T>;
    /**
     * {@link https://dom.spec.whatwg.org/#dom-element-closest}
     */
    closest(predicate: Predicate<Node<F>>, options?: Flags<F>): Option<Node<F>>;
    [Symbol.iterator](): Iterator<Node<F>>;
    equals(value: Node<F>): boolean;
    equals(value: unknown): value is this;
    hash(hash: Hash): void;
    toJSON(options?: json.Serializable.Options): Node.JSON<T>;
    /**
     * @internal
     */
    _attachParent(parent: Node<F>): boolean;
}
/**
 * @public
 */
export declare namespace Node {
    interface JSON<T extends string = string> {
        [key: string]: json.JSON | undefined;
        type: T;
        children?: Array<JSON>;
        externalId?: string;
        serializationId?: string;
    }
}
//# sourceMappingURL=tree.d.ts.map